#include <stdio.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <sys/ioctl.h> 
#include <netinet/in.h>
#include <netinet/tcp.h> 
#include <netdb.h>
#include <errno.h>
#include <glib.h>
#include <unistd.h>
#include <glib.h>
#include <libgnomevfs/gnome-vfs.h>

#include "common.h"


static int
destroy_tcp_connection(tcp_con_t *con) {

  dbg_out("Here\n");
  if (!con)
    return -EINVAL;

  if (con->self)
    freeaddrinfo(con->self);

  close(con->soc);

  return 0;
}

static int
parse_request(request_msg_t *req,const char *request_string){
  size_t len;
  long int_val;
  char *sp=NULL;
  char *ep=NULL;
  char *buffer;
  ssize_t remains;
  gboolean has_offset=FALSE;
  int rc=0;

  if  ( (!request_string) || (!req) )
    return -EINVAL;

  dbg_out("Here:%s\n",request_string);

  buffer=strdup(request_string);
  if (!buffer)
    return -ENOMEM;

  len=strlen(request_string);
  _assert(len>0);
  remains=len;
  /*
   * パケット番号
   */
  sp=buffer;
  ep=memchr(sp, ':', remains);
  if (!ep) {
    rc=-EINVAL;
    goto err_out;
  }

  *ep='\0';
  remains =len - ((unsigned long)ep-(unsigned long)buffer);
  if (remains<=0) {
    rc=-EINVAL;
    goto err_out;
  }
  ++ep;
  int_val=strtol(sp, (char **)NULL, 16);
  req->pkt_no=int_val;
  dbg_out("pktno:%ld\n",req->pkt_no);
  sp=ep;

  /*
   * ファイルID
   */
  ep=memchr(sp, ':', remains);
  if (ep) {
    has_offset=TRUE;
    *ep='\0';
    remains =len - ((unsigned long)ep-(unsigned long)buffer);
    if (remains<=0) {
      rc=-EINVAL;
      goto err_out;
    }
    ++ep;
  }
  int_val=strtol(sp, (char **)NULL, 16);
  req->fileid=int_val;
  dbg_out("fileid:%d(%x)\n",req->fileid,req->fileid);
  sp=ep;

  /*
   * オフセット
   */
  if (has_offset)
    int_val=strtol(sp, (char **)NULL, 16);
  else
    int_val=0; /* 仮に0にする */
  req->offset=int_val;
  dbg_out("offset:%d(%x)\n",req->offset,req->offset);
err_out:
  free(buffer);
  return rc;
}
static int
tcp_flush_buff_size(tcp_con_t *con){
  int rc;
  int size=0;

  /* 念のため書き込みソケットバッファ長を0に設定 */
  size=0;
  rc=setsockopt(con->soc, SOL_SOCKET, SO_SNDBUF, (void*)&size, sizeof(int));
  if (rc<0) {
    err_out("Can not set buffer size:%s(%d)\n",strerror(errno),errno);
  }
  /* 念のため書き込みソケットバッファ長を0に設定 */
  size=0;
  rc=setsockopt(con->soc, SOL_SOCKET, SO_RCVBUF, (void*)&size, sizeof(int));
  if (rc<0) {
    err_out("Can not set buffer size:%s(%d)\n",strerror(errno),errno);
  }

  return 0;
}
static int
tcp_ipmsg_finalize_connection(tcp_con_t *con){
  int rc=0;
  char dummy[1];

  g_assert(con);
  tcp_flush_buff_size(con);
  tcp_flush_buffer(con);
  sock_set_buffer(con->soc);
  dbg_out("Wait for closing by peer.\n");
  sock_set_recv_timeout(con->soc,TCP_CLOSE_WAIT_SEC);
  rc=wait_socket(con->soc,WAIT_FOR_READ,TCP_SELECT_SEC);
  if (rc<0) {
    dbg_out("OK timeout select :%s(%d)\n",strerror(errno),errno);
    return -errno;
  }
  memset(dummy, 0, sizeof(dummy));
  rc=recv(con->soc, dummy,sizeof(dummy),0); /* 相手のクローズ検出 */

  if (rc<0)
    dbg_out("OK timeout:%s(%d)\n",strerror(errno),errno);
  else
    rc=0;

  return rc;
}
static int
tcp_transfer_file(tcp_con_t *con,const char *path,const size_t size,off_t offset){
  int fd;
  int rc;
  char buff[TCP_FILE_BUFSIZ];
  ssize_t read_len;
  ssize_t file_remains;
  ssize_t soc_remains;
  size_t write_len;
  size_t total_write;
  char *wp;
  struct sigaction saved_act;

  if ( (!con) || (!path) )
    return -EINVAL;

  fd=open(path,O_RDONLY);
  if (fd<0)
    return -errno;

  rc=lseek(fd, offset, SEEK_SET);
  if (rc<0) {
    rc=-errno;
    goto close_out;
  }
  total_write=0;
  file_remains=size;

  while(file_remains>0) {
    read_len=read(fd,buff,TCP_FILE_BUFSIZ);
    if (read_len<0) {
      rc=-errno;
      goto close_out;
    }
    file_remains -= read_len;
    soc_remains=read_len;
    wp=buff;
    if (wait_socket(con->soc,WAIT_FOR_WRITE,TCP_SELECT_SEC)<0) {
      err_out("Can not send socket\n");
      goto close_out;
    }
    while(soc_remains>0) {
      disable_pipe_signal(&saved_act);
      write_len=send(con->soc,wp,soc_remains,0);
      enable_pipe_signal(&saved_act);
      if (write_len<0) {
	if (errno==EINTR)
	  continue;
	goto close_out;
      }
      wp += write_len;
      total_write += write_len;
      soc_remains -= write_len;
    }
    dbg_out("transfer %s %d/%d(%d)\n",path,total_write,size,file_remains);
  }
  tcp_flush_buffer(con);
  rc=0;
 close_out:
  close(fd);
  if (rc<0)
    dbg_out("Can not send file:%s %s %d\n",path,strerror(errno),errno);
  return rc;
}
static int
create_response(unsigned long type,const GnomeVFSFileInfo *info,const char *dir,char **response){
  int rc=0;
  gchar *fullpath;
  char buff[IPMSG_BUFSIZ];
  size_t act_len;
  gchar *res;
  char *name;

  if ( (!info) || (!dir) || (!response) )
    return -EINVAL;

  fullpath=g_build_filename(dir,info->name,NULL);
  if (!fullpath)
    return -ENOMEM;
  /* FIXME
   * きれいにすること
   */
  dbg_out("file:%s size:%d(%x)\n",
	  info->name,info->size,info->size);
  switch(type)
    {
    case IPMSG_FILE_REGULAR:
      snprintf(buff,IPMSG_BUFSIZ-1,"%04x:%s:%x:1:",
	       0,info->name,info->size);
      buff[IPMSG_BUFSIZ-1]='\0';
      act_len=strlen(buff);
      snprintf(buff,IPMSG_BUFSIZ-1,"%04x:%s:%x:1:",
	       act_len,info->name,info->size);
      buff[IPMSG_BUFSIZ-1]='\0';
      break;
    case IPMSG_FILE_DIR:
      snprintf(buff,IPMSG_BUFSIZ-1,"%04x:%s:0:2:",
	       0,info->name);
      buff[IPMSG_BUFSIZ-1]='\0';
      act_len=strlen(buff);
      snprintf(buff,IPMSG_BUFSIZ-1,"%04x:%s:0:2:",
	       act_len,info->name);
      buff[IPMSG_BUFSIZ-1]='\0';
      break;
    case IPMSG_FILE_RETPARENT:
      snprintf(buff,IPMSG_BUFSIZ-1,"%04x:.:0:3:",
	       0);
      buff[IPMSG_BUFSIZ-1]='\0';
      act_len=strlen(buff);
      snprintf(buff,IPMSG_BUFSIZ-1,"%04x:.:0:3:",
	       act_len);
      buff[IPMSG_BUFSIZ-1]='\0';
      break;
    default:
      rc=-EINVAL;
      break;
  }
  rc=convert_string_ipmsg_proto(buff,&res);
  if (rc<0) {
    err_out("Can not convert code set:%s\n",buff);
    return -ENOMEM;
  }
  *response=res;

  return rc;
}
static int
send_header(tcp_con_t *con,const char *response){
  int rc;
  size_t soc_remains;
  ssize_t send_len;
  char *wp;
  int buff_len;
  struct sigaction saved_act;

  if ( (!con) || (!response) )
    return -EINVAL;

  soc_remains=strlen(response);
  if (wait_socket(con->soc,WAIT_FOR_WRITE,TCP_SELECT_SEC)<0) {
    err_out("Can not send socket\n");
    goto error_out;
  }

  wp=(char *)response;
  dbg_out("Send:%s(%d, %x)\n",wp,soc_remains,soc_remains);
  while(soc_remains>0) {
    disable_pipe_signal(&saved_act);
    send_len=send(con->soc,wp,soc_remains,0);
    enable_pipe_signal(&saved_act);
    if (send_len<0) {
      if (errno==EINTR)
	continue;
      rc=-errno;
      err_out("Error:%s (%d)\n",strerror(errno),errno);
      goto error_out;
    }
    wp += send_len;
    soc_remains -= send_len;
    dbg_out("soc remains:%d\n",soc_remains);
  }
  rc=0;
 error_out:
  tcp_flush_buffer(con);
  return rc; 
}
static int
read_directory_files(tcp_con_t *con,const char *parent,GnomeVFSFileInfo *info) {
  int rc;
  char *rel_path;
  char *full_path;
  char *res_message;

  if ( (!con) || (!parent) || (!info) )
    return -EINVAL;

  if ( (!strcmp("..",info->name)) || (!strcmp(".",info->name) ) )
    return -EINVAL;

  if (info->type==GNOME_VFS_FILE_TYPE_REGULAR) {
    rc=create_response(IPMSG_FILE_REGULAR,info,parent,&res_message);
    if (rc<0)
      return rc;
    dbg_out("Send file:%s\n",res_message);
    rc=send_header(con,res_message);
    free(res_message);
    if (rc<0)
      return rc;
  }

  rc=-ENOMEM;
  full_path=g_build_filename(parent,info->name,NULL);
  if (!full_path) 
    return rc;
  /* send file */
  tcp_transfer_file(con,full_path,info->size,0);
  g_free(full_path);
  return 0;
}
static int
send_directory(tcp_con_t *con,const char *top_dir,const char *basename,GnomeVFSFileInfo *info){
  int rc;
  char *uri;
  GnomeVFSResult res;
  GnomeVFSDirectoryHandle *handle;
  char *next_dir;
  char *res_message;

  if ( (!top_dir) || (!info) || (!basename) )
    return -EINVAL;

  rc=create_response(IPMSG_FILE_DIR,info,top_dir,&res_message);
  
  if (rc<0)
    return rc;
  dbg_out("Send dir:%s (%s)\n",res_message,top_dir);
  rc=send_header(con,res_message);
  free(res_message);
  if (rc<0)
    return rc;

      
  uri=gnome_vfs_get_uri_from_local_path(top_dir);
  if (!uri)
    return -ENOMEM;
  /*
   *ファイルを送付
   */
  res=gnome_vfs_directory_open(&handle,uri,GNOME_VFS_FILE_INFO_FOLLOW_LINKS);
  res=gnome_vfs_directory_read_next(handle,info);
  while (res==GNOME_VFS_OK) {
    read_directory_files(con,top_dir,info);
    res=gnome_vfs_directory_read_next(handle,info);
  }
  gnome_vfs_directory_close(handle); 
  /*
   *ディレクトリを送付
   */
  res=gnome_vfs_directory_open(&handle,uri,GNOME_VFS_FILE_INFO_FOLLOW_LINKS);
  if (res!=GNOME_VFS_OK) {
    err_out("Can not open dir:%s %s (%d)\n",
	    top_dir,
	    gnome_vfs_result_to_string(res),
	    res);
    goto error_out;
  }
  res=gnome_vfs_directory_read_next(handle,info);
  if (res!=GNOME_VFS_OK) {
    if (res != GNOME_VFS_ERROR_EOF)
      err_out("Can not read next dir:%s %s (%d)\n",
	      top_dir,
	      gnome_vfs_result_to_string(res),
	      res);
    goto error_out;
  }
  while (res==GNOME_VFS_OK) {
    if ( (info->type==GNOME_VFS_FILE_TYPE_DIRECTORY) &&
	 ( (strcmp(info->name,"..")) && (strcmp(info->name,"."))) ){
      dbg_out("dir:%s\n",info->name);
      next_dir=g_build_filename(top_dir,info->name,NULL);
      if (!next_dir)
	goto error_out;
      rc=send_directory(con,next_dir,info->name,info);
      g_free(next_dir);
      if (rc<0)
	return rc;
      rc=create_response(IPMSG_FILE_RETPARENT,info,top_dir,&res_message);
      if (rc<0)
	return rc;
      dbg_out("Send ret parent:%s\n",res_message);
      rc=send_header(con,res_message);
      free(res_message);
      if (rc<0)
	return rc;
    }
    res=gnome_vfs_directory_read_next(handle,info);
    if (res!=GNOME_VFS_OK) {
      if (res != GNOME_VFS_ERROR_EOF)
	err_out("Can not read next dir:%s %s (%d)\n",
		top_dir,
		gnome_vfs_result_to_string(res),
		res);
      goto error_out;
    }
  }

 error_out:
  gnome_vfs_directory_close(handle); 
  g_free(uri);
  return rc;
}
static int
finalize_send_directory(tcp_con_t *con,const char *top_dir,GnomeVFSFileInfo *info){
  int rc;
  char *res_message;

  if ( (!top_dir) || (!info) )
    return -EINVAL;

  rc=create_response(IPMSG_FILE_RETPARENT,info,top_dir,&res_message);
  if (rc<0)
    return rc;
  dbg_out("Send return to top dir:%s (%s)\n",res_message,top_dir);
  rc=send_header(con,res_message);
  free(res_message);

  return rc;
}


static int
tcp_transfer_dir(tcp_con_t *con,const char *path){
  int rc;
  char *basename;
  GnomeVFSFileInfo *info;
  GnomeVFSResult res;
  gchar *dir_uri;

  if ( (!con) || (!path) )
    return -EINVAL;
  info=gnome_vfs_file_info_new();
  dir_uri=gnome_vfs_get_uri_from_local_path(path);

  rc=-ENOMEM;
  if (!dir_uri) 
    goto free_info_out;

  res=gnome_vfs_get_file_info(dir_uri,info,GNOME_VFS_FILE_INFO_FOLLOW_LINKS);
  if (res != GNOME_VFS_OK) {
    rc=-res;
    goto free_info_out;
  }
  basename=g_path_get_basename(path);
    rc=-ENOMEM;
  if (!basename)
    goto free_info_out;

  rc=send_directory(con,path,basename,info);
  g_free(basename);
  if (!rc)
    rc=finalize_send_directory(con,path,info);
 free_info_out:
  gnome_vfs_file_info_unref(info);
  return rc;
}
int
disable_pipe_signal(struct sigaction *saved_act){
  int rc;
  struct sigaction new_act;
  
  if (!saved_act)
    return -EINVAL;

  memset(&new_act,0,sizeof(struct sigaction));

  new_act.sa_handler=SIG_IGN;
  rc=sigaction(SIGPIPE,&new_act,saved_act);
  g_assert(!rc);
  
  return rc;
}
int
enable_pipe_signal(struct sigaction *saved_act){
  int rc;
  
  if (!saved_act)
    return -EINVAL;

  rc=sigaction(SIGPIPE,saved_act,NULL);
  g_assert(!rc);
  
  return rc;
  
}
#if defined(TIOCOUTQ) && defined(TIOCINQ)
int
tcp_flush_buffer(tcp_con_t *con){
  int rc;
  int nodelay=1;
  int w_buff_len;
  int r_buff_len;
  int max_count=TCP_FLUSH_RETRY;
  if (!con)
    return -EINVAL;

  tcp_flush_buff_size(con);
  do{
    ipmsg_update_ui();

    rc=ioctl(con->soc,TIOCOUTQ,&w_buff_len);
    dbg_out("TCPWriteBuffRemains:%d\n",w_buff_len);
    if (rc<0) {
      dbg_out("Can not obtain remained data in the tcp buffer %s %d\n",strerror(errno),errno);
      w_buff_len=0;
    }
    rc=ioctl(con->soc,TIOCINQ,&r_buff_len);
    dbg_out("TCPReadBuffRemains:%d\n",r_buff_len);
    if (rc<0) {
      dbg_out("Can not obtain remained data in the tcp buffer %s %d\n",strerror(errno),errno);
      r_buff_len=0;
    }
    rc=setsockopt(con->soc, IPPROTO_TCP, TCP_NODELAY, (void*)&nodelay, sizeof(int));
    if (rc<0) {
      err_out("Can not flush buffer:%s(%d)\n",strerror(errno),errno);
      return -errno;
    }
    --max_count;
  }while( (max_count>0) &&( (w_buff_len) || (r_buff_len) ) );
  sock_set_buffer(con->soc);

  return 0;
}
#else /* For FreeBSD */
int
tcp_flush_buffer(tcp_con_t *con){
  if (!con)
    return -EINVAL;

  tcp_flush_buff_size(con);
  ipmsg_update_ui();
  sock_set_buffer(con->soc);

  return 0;
}
#endif
int
tcp_enable_keepalive(const tcp_con_t *con){
  int rc;
  int flag;

  if (!con) 
    return -EINVAL;

  flag=1;
  rc=setsockopt(con->soc, SOL_SOCKET, SO_KEEPALIVE, (void*)&flag, sizeof(int));
  if (rc<0) {
    err_out("Can not set keepalive:%s(%d)\n",strerror(errno),errno);
    return -errno;
  }

  return 0;
}

int
tcp_setup_client(int family,const char *ipaddr,int port,tcp_con_t *con){
  int rc;
  int sock=-1;
  struct sockaddr_in server;
  struct hostent *host;
  struct addrinfo *info=NULL;

  if ( (!ipaddr) || (!con) )
    return -EINVAL;

  rc=setup_addr_info(&info, ipaddr,port,SOCK_STREAM,family);
  if (rc<0)
    return rc;

  sock=socket(info->ai_family, info->ai_socktype,info->ai_protocol);

  if (sock<0) {
    err_out("Can not create socket: %s(errno:%d)\n",strerror(errno),errno);
    rc=-errno;
    goto error_out;    
  }

#ifdef IPV6_V6ONLY
  if (info->ai_family == AF_INET6) {
    int v6only=1;

    rc=setsockopt(sock,IPPROTO_IPV6,IPV6_V6ONLY,(void *)&v6only,sizeof(v6only));
    if (rc) {
      goto error_out;
    }
  }
#endif /*  IPV6_V6ONLY  */

  if (connect(sock, info->ai_addr, info->ai_addrlen)<0){
    err_out("Can not connect addr:%s port:%d %s (errno:%d)\n",
	    ipaddr,
	    port,
	    strerror(errno),
	    errno);
    rc=-errno;
    goto error_out;    
  }
  rc=sock_set_buffer(sock);
  if (rc<0) {
    err_out("Can not set socket buffer: %s (%d)\n",strerror(errno),errno);
    return -errno;
  }
  /* 受信時はタイムアウト待ちを行うのでここでは, ノンブロックにしないこと  */
  con->soc=sock;
  con->family=info->ai_family;
  con->self=info;

  return 0;
 error_out:
  if (info)
    freeaddrinfo(info);
  return rc;
}

int
tcp_set_buffer(tcp_con_t *con) {
  if (!con)
    return -EINVAL;

 return sock_set_buffer(con->soc);
}
gpointer
ipmsg_tcp_recv_thread(gpointer data){
  ssize_t recv_len;
  size_t addr_len;
  int rc;
  tcp_con_t *con;
  char recv_buf[_MSG_BUF_SIZE];
  int count=200;
  struct addrinfo *info;

  con=(tcp_con_t *)data;
  if (!con) {
    err_out("No connection recived\n");
    g_assert(con);
    return NULL;
  }
  rc=tcp_enable_keepalive(con);
  if (rc<0)
    return NULL;
  
#if 0
  rc=tcp_set_buffer(con);
  if (rc<0) {
    return NULL;
  }
#endif
  while(1) {
    if (wait_socket(con->soc,WAIT_FOR_READ,TCP_SELECT_SEC)<0) {
      err_out("Can not send socket\n");
      return NULL;
    }else{
    read_retry:
      errno=0;
      memset(recv_buf, 0, sizeof(recv_buf));
      info=con->self;
      recv_len=recv(con->soc,
		    recv_buf,
		    _MSG_BUF_SIZE,
		    MSG_PEEK);

      if (recv_len<=0) {
	if (errno==EINTR)
	  goto read_retry;
	if (errno)
	  err_out("Can not peek message %s(errno:%d)\n",strerror(errno),errno);
	destroy_tcp_connection(con);
	free(con);
	return NULL;
      }

      dbg_out("tcp peek read:%d %s\n",recv_len,recv_buf);

      recv_len=recv(con->soc,
		    recv_buf,
		    recv_len,
		    MSG_PEEK);


      dbg_out("tcp read:%d %s\n",recv_len,recv_buf);
      if (recv_len>0) {
	msg_data_t msg;
	request_msg_t req;
	char *path;
	size_t size;
	unsigned long ipmsg_fattr;

	recv_buf[recv_len-1]='\0';
	memset(&req,0,sizeof(request_msg_t));

	init_message_data(&msg);
	parse_message(&msg,recv_buf,recv_len);
	parse_request(&req,msg.message);
	rc=refer_attach_file(req.pkt_no,req.fileid,&ipmsg_fattr,(const char **)&path,&size);
	if (rc<0) {
	  err_out("Can not find message:pktno %ld id : %d\n",
		  req.pkt_no,
		  req.fileid);
	  close(con->soc); /* エラーとしてクローズする  */
	}else{
	  dbg_out("transfer:%s (%d)\n",path,size);
	  switch(ipmsg_fattr) 
	    {
	    case IPMSG_FILE_REGULAR:
	      if (!tcp_transfer_file(con,path,size,0)) {
		tcp_ipmsg_finalize_connection(con);
		download_monitor_release_file(req.pkt_no,req.fileid);
	      }
	      break;
	    case IPMSG_FILE_DIR:
	      if (!tcp_transfer_dir(con,path)){
		download_monitor_release_file(req.pkt_no,req.fileid);
		tcp_ipmsg_finalize_connection(con);
	      }
	      break;
	    default:
	      break;
	    }
	  release_message_data(&msg);
	  free(path);
	  break;
	}
	release_message_data(&msg);
	break;
      }
    }
    --count;
    if (!count)
      break;
  }
  if (con->self) /* closeは相手側で行うので, destroy_tcp_connectionは呼び出せない
		  * (Win版ipmsgの仕様).  
		  */
    freeaddrinfo(con->self); 
  free(con);

  return NULL;
}
gpointer 
ipmsg_tcp_server_thread(gpointer data){
  int rc;
  size_t len;
  int sock;
  int tcp_socket;
  int port;
  int reuse;
  int family;
  tcp_con_t *client_con=NULL;
  struct addrinfo *info=NULL;

  family=(int)data;
  rc=setup_addr_info(&info, 
		     NULL,
		     hostinfo_refer_ipmsg_port(),
		     SOCK_STREAM,family);
  if (rc<0)
    return NULL;

  tcp_socket = socket(info->ai_family,
		      info->ai_socktype,
		      info->ai_protocol);

  if (tcp_socket < 0) {
    err_out("Can not create socket:%s (%d)\n",strerror(errno),errno);
    goto error_out;
  }

  reuse=1;
  if (setsockopt(tcp_socket,SOL_SOCKET, SO_REUSEADDR, (const char *)&reuse, sizeof(reuse))<0) {
    err_out("Can not set sock opt:%s (%d)\n",strerror(errno),errno);
    goto error_out;
  }

#ifdef IPV6_V6ONLY
  if (info->ai_family == AF_INET6) {
    int v6only=1;

    rc=setsockopt(tcp_socket,IPPROTO_IPV6,IPV6_V6ONLY,(void *)&v6only,sizeof(v6only));
    if (rc) {
      goto error_out;
    }
  }
#endif /*  IPV6_V6ONLY  */

  if (bind(tcp_socket, info->ai_addr, info->ai_addrlen) != 0) {
    err_out("Can not bind socket:%s (%d)\n",strerror(errno),errno);
    goto error_out;
  }

  if (listen(tcp_socket, TCP_LISTEN_LEN) != 0) {
    err_out("Can not listen socket:%s (%d)\n",strerror(errno),errno);
    goto error_out;
  }

  if (sock_set_buffer(tcp_socket)<0) {
    err_out("Can not set socket buffer:%s (%d)\n",strerror(errno),errno);
    goto error_out;
  }

  while (1) {
    GThread *handler;
    struct addrinfo *client_info=NULL;

    client_con=malloc(sizeof(tcp_con_t));
    g_assert(client_con);
    dbg_out("accept\n");

    
    rc=setup_addr_info(&client_info, 
		     NULL,
		     hostinfo_refer_ipmsg_port(),
		     SOCK_STREAM,family);
    
    if (rc<0)
      goto free_client_out;
    
    sock = accept(tcp_socket, client_info->ai_addr,&(client_info->ai_addrlen));
    if (sock < 0) {
      err_out("Can not accept:%s (%d)\n",strerror(errno),errno);
      goto free_client_out;
    }
    if (sock_set_buffer(sock)<0) {
      err_out("Can not set socket buffer:%s (%d)\n",strerror(errno),errno);
      goto free_client_out;
    }

    client_con->soc=sock;
    client_con->self=client_info;
    handler=g_thread_create(ipmsg_tcp_recv_thread,
                            client_con,
			    FALSE,
			    NULL);
    if (!handler) {
      err_out("Can not create handler thread.\n");
      goto free_client_out;
    }
  } 
  return NULL;

 free_client_out:  
  destroy_tcp_connection(client_con);
 error_out:
  if (info)
    freeaddrinfo(info);
  exit(0); /* 多重起動のため 終了  */
  return NULL;
}
